from typing import Type, Dict, Any, Optional, Union, Callable
from pathlib import Path
import importlib
import logging
import torch
import torch.nn as nn
from threading import Lock

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class Registry:
    _instance = None
    _lock = Lock()
    
    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._instance._initialized = False
            return cls._instance
    
    def __init__(self):
        if not self._initialized:
            self._registry = {
                'model': {},
                'activation': {},
                'normalization': {},
                'block': {},
                'attention': {},
            }
            self._initialized = True
    
    def register(self, component_type: str, name: str, cls: Type[nn.Module], **kwargs):
        if component_type not in self._registry:
            raise ValueError(
                f"Unknown component type: {component_type}. "
                f"Available types: {list(self._registry.keys())}"
            )
        self._registry[component_type][name] = {
            'class': cls,
            'default_args': kwargs
        }
        
    def get(self, component_type: str, name: str) -> Dict[str, Any]:
        if component_type not in self._registry:
            raise ValueError(f"Unknown component type: {component_type}")
        name = name.lower()
        if name not in self._registry[component_type]:
            raise ValueError(
                f"Unknown {component_type}: {name}. "
                f"Available {component_type}s: {self.list_components(component_type)}"
            )
        return self._registry[component_type][name]
    
    def list_components(self, component_type: Optional[str] = None) -> Union[list, Dict[str, list]]:
        if component_type:
            if component_type not in self._registry:
                raise ValueError(f"Unknown component type: {component_type}")
            return sorted(self._registry[component_type].keys())
        return {k: sorted(v.keys()) for k, v in self._registry.items()}

registry = Registry()

def register_component(component_type: str, name: str = None, **kwargs):
    def wrapper(cls: Type[nn.Module]):
        component_name = (name or cls.__name__).lower()
        registry.register(component_type, component_name, cls, **kwargs)
        logger.info(f"Registered {component_type}: {component_name}")
        return cls
    return wrapper

def register_model(name: str = None, **kwargs):
    return register_component('model', name, **kwargs)

def create_component(component_type: str, name: str, **kwargs) -> nn.Module:
    component_info = registry.get(component_type, name)
    cls = component_info['class']
    args = component_info['default_args'].copy()
    args.update(kwargs)
    try:
        return cls(**args)
    except Exception as e:
        raise RuntimeError(f"Error creating {component_type} '{name}': {str(e)}")

def build_model(
    model_name: str,
    config: Optional[Union[str, dict]] = None,
    device: Optional[str] = None,
    **kwargs
) -> nn.Module:
    model_name = model_name.lower()
    final_config = {}
    try:
        model_info = registry.get('model', model_name)
        final_config.update(model_info['default_args'])
    except ValueError:
        pass
    
    if config:
        if isinstance(config, str):
            try:
                config_module = importlib.import_module(config)
                final_config.update(vars(config_module))
            except ImportError:
                raise ImportError(f"Could not import config from {config}")
        else:
            final_config.update(config)
    
    final_config.update(kwargs)
    model = create_component('model', model_name, **final_config)
    if device:
        model = model.to(device)
    return model

def list_components(component_type: Optional[str] = None) -> Union[list, Dict[str, list]]:
    return registry.list_components(component_type)

def list_models():
    return list_components('model')

def save_model(model: nn.Module, save_path: Union[str, Path], save_config: bool = True):
    save_path = Path(save_path)
    save_path.parent.mkdir(exist_ok=True, parents=True)
    torch.save(model.state_dict(), save_path)
    if save_config and hasattr(model, 'config'):
        import json
        config_path = save_path.with_suffix('.json')
        with open(config_path, 'w') as f:
            json.dump(model.config, f, indent=4)

def load_model(
    model_name: str,
    weights_path: Union[str, Path],
    config: Optional[Union[str, dict]] = None,
    device: Optional[str] = None,
    **kwargs
) -> nn.Module:
    model = build_model(model_name, config, device, **kwargs)
    state_dict = torch.load(weights_path, map_location=device or 'cpu')
    model.load_state_dict(state_dict)
    return model.to(device or 'cpu')